@@ -72,6 +72,8 @@ static char *_dl_debug_detail  = 0;
 static char *_dl_debug_nofixups  = 0;
 static char *_dl_debug_bindings  = 0;
 static int   _dl_debug_file = 2;
+#elif defined __SUPPORT_LD_DEBUG_EARLY__
+#define _dl_debug_file 2
 #endif
 char *_dl_library_path = 0;
 char *_dl_ldsopath = 0;
@@ -133,7 +135,7 @@ void *_dlopen(const char *libname, int flag)
 #endif
 
 	/* A bit of sanity checking... */
-	if (!(flag & RTLD_LAZY|RTLD_NOW)) {
+	if (!(flag & (RTLD_LAZY|RTLD_NOW))) {
 		_dl_error_number = LD_BAD_HANDLE;
 		return NULL;
 	}
