@@ -45,6 +45,16 @@
  * 2-Feb-2002 Erik Andersen <andersee@debian.org>
  * Added gethostent(), sethostent(), and endhostent()
  *
+ * 17-Aug-2002 Manuel Novoa III <mjn3@codepoet.org>
+ *   Fixed __read_etc_hosts_r to return alias list, and modified buffer
+ *   allocation accordingly.  See MAX_ALIASES and ALIAS_DIM below.
+ *   This fixes the segfault in the Python 2.2.1 socket test.
+ *
+ * 04-Jan-2003 Jay Kulpinski <jskulpin@berkshire.rr.com>
+ *   Fixed __decode_dotted to count the terminating null character
+ *   in a host name.
+ *
+ *
  */
 
 #define __FORCE_GLIBC
@@ -55,6 +65,7 @@
 #include <errno.h>
 #include <sys/socket.h>
 #include <sys/types.h>
+#include <sys/time.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
 #include <stdlib.h>
@@ -72,6 +83,11 @@
 #define MAX_SERVERS 3
 #define MAX_SEARCH 4
 
+#define MAX_ALIASES	5
+
+/* 1:ip + 1:full + MAX_ALIASES:aliases + 1:NULL */
+#define 	ALIAS_DIM		(2 + MAX_ALIASES + 1)
+
 #undef DEBUG
 /*#define DEBUG*/
 
@@ -82,11 +98,21 @@
 #endif /* DEBUG */
 
 
-/* Global stuff... */
-extern int nameservers;
-extern char * nameserver[MAX_SERVERS];
-extern int searchdomains;
-extern char * searchdomain[MAX_SEARCH];
+/* Global stuff (stuff needing to be locked to be thread safe)... */
+extern int __nameservers;
+extern char * __nameserver[MAX_SERVERS];
+extern int __searchdomains;
+extern char * __searchdomain[MAX_SEARCH];
+
+#ifdef __UCLIBC_HAS_THREADS__
+#include <pthread.h>
+extern pthread_mutex_t __resolv_lock;
+# define BIGLOCK	pthread_mutex_lock(&__resolv_lock)
+# define BIGUNLOCK	pthread_mutex_unlock(&__resolv_lock);
+#else
+# define BIGLOCK
+# define BIGUNLOCK
+#endif
 
 
 
@@ -123,46 +149,47 @@ enum etc_hosts_action {
 };
 
 /* function prototypes */
-extern int get_hosts_byname_r(const char * name, int type,
+extern int __get_hosts_byname_r(const char * name, int type,
 			      struct hostent * result_buf,
 			      char * buf, size_t buflen,
 			      struct hostent ** result,
 			      int * h_errnop);
-extern int get_hosts_byaddr_r(const char * addr, int len, int type,
+extern int __get_hosts_byaddr_r(const char * addr, int len, int type,
 			      struct hostent * result_buf,
 			      char * buf, size_t buflen,
 			      struct hostent ** result,
 			      int * h_errnop);
 extern void __open_etc_hosts(FILE **fp);
-extern int read_etc_hosts_r(FILE *fp, const char * name, int type,
+extern int __read_etc_hosts_r(FILE *fp, const char * name, int type,
 			    enum etc_hosts_action action,
 			    struct hostent * result_buf,
 			    char * buf, size_t buflen,
 			    struct hostent ** result,
 			    int * h_errnop);
-extern int dns_lookup(const char * name, int type, int nscount, 
+extern int __dns_lookup(const char * name, int type, int nscount, 
 	char ** nsip, unsigned char ** outpacket, struct resolv_answer * a);
 
-extern int encode_dotted(const char * dotted, unsigned char * dest, int maxlen);
-extern int decode_dotted(const unsigned char * message, int offset, 
+extern int __encode_dotted(const char * dotted, unsigned char * dest, int maxlen);
+extern int __decode_dotted(const unsigned char * message, int offset, 
 	char * dest, int maxlen);
-extern int length_dotted(const unsigned char * message, int offset);
-extern int encode_header(struct resolv_header * h, unsigned char * dest, int maxlen);
-extern int decode_header(unsigned char * data, struct resolv_header * h);
-extern int encode_question(struct resolv_question * q,
+extern int __length_dotted(const unsigned char * message, int offset);
+extern int __encode_header(struct resolv_header * h, unsigned char * dest, int maxlen);
+extern int __decode_header(unsigned char * data, struct resolv_header * h);
+extern int __encode_question(struct resolv_question * q,
 	unsigned char * dest, int maxlen);
-extern int decode_question(unsigned char * message, int offset,
+extern int __decode_question(unsigned char * message, int offset,
 	struct resolv_question * q);
-extern int encode_answer(struct resolv_answer * a,
+extern int __encode_answer(struct resolv_answer * a,
 	unsigned char * dest, int maxlen);
-extern int decode_answer(unsigned char * message, int offset,
+extern int __decode_answer(unsigned char * message, int offset,
 	struct resolv_answer * a);
-extern int length_question(unsigned char * message, int offset);
-extern int open_nameservers(void);
+extern int __length_question(unsigned char * message, int offset);
+extern int __open_nameservers(void);
+extern void __close_nameservers(void);
 
 
 #ifdef L_encodeh
-int encode_header(struct resolv_header *h, unsigned char *dest, int maxlen)
+int __encode_header(struct resolv_header *h, unsigned char *dest, int maxlen)
 {
 	if (maxlen < HFIXEDSZ)
 		return -1;
@@ -189,7 +216,7 @@ int encode_header(struct resolv_header *h, unsigned char *dest, int maxlen)
 #endif
 
 #ifdef L_decodeh
-int decode_header(unsigned char *data, struct resolv_header *h)
+int __decode_header(unsigned char *data, struct resolv_header *h)
 {
 	h->id = (data[0] << 8) | data[1];
 	h->qr = (data[2] & 0x80) ? 1 : 0;
@@ -213,7 +240,7 @@ int decode_header(unsigned char *data, struct resolv_header *h)
    This routine is fairly dumb, and doesn't attempt to compress
    the data */
 
-int encode_dotted(const char *dotted, unsigned char *dest, int maxlen)
+int __encode_dotted(const char *dotted, unsigned char *dest, int maxlen)
 {
 	int used = 0;
 
@@ -247,7 +274,7 @@ int encode_dotted(const char *dotted, unsigned char *dest, int maxlen)
 /* Decode a dotted string from nameserver transport-level encoding.
    This routine understands compressed data. */
 
-int decode_dotted(const unsigned char *data, int offset,
+int __decode_dotted(const unsigned char *data, int offset,
 				  char *dest, int maxlen)
 {
 	int l;
@@ -285,6 +312,11 @@ int decode_dotted(const unsigned char *data, int offset,
 			dest[used++] = '\0';
 	}
 
+	/* The null byte must be counted too */
+	if (measure) {
+	    total++;
+	}
+
 	DPRINTF("Total decode len = %d\n", total);
 
 	return total;
@@ -293,7 +325,7 @@ int decode_dotted(const unsigned char *data, int offset,
 
 #ifdef L_lengthd
 
-int length_dotted(const unsigned char *data, int offset)
+int __length_dotted(const unsigned char *data, int offset)
 {
 	int orig_offset = offset;
 	int l;
@@ -316,12 +348,12 @@ int length_dotted(const unsigned char *data, int offset)
 #endif
 
 #ifdef L_encodeq
-int encode_question(struct resolv_question *q,
+int __encode_question(struct resolv_question *q,
 					unsigned char *dest, int maxlen)
 {
 	int i;
 
-	i = encode_dotted(q->dotted, dest, maxlen);
+	i = __encode_dotted(q->dotted, dest, maxlen);
 	if (i < 0)
 		return i;
 
@@ -341,13 +373,13 @@ int encode_question(struct resolv_question *q,
 #endif
 
 #ifdef L_decodeq
-int decode_question(unsigned char *message, int offset,
+int __decode_question(unsigned char *message, int offset,
 					struct resolv_question *q)
 {
 	char temp[256];
 	int i;
 
-	i = decode_dotted(message, offset, temp, sizeof(temp));
+	i = __decode_dotted(message, offset, temp, sizeof(temp));
 	if (i < 0)
 		return i;
 
@@ -362,11 +394,11 @@ int decode_question(unsigned char *message, int offset,
 #endif
 
 #ifdef L_lengthq
-int length_question(unsigned char *message, int offset)
+int __length_question(unsigned char *message, int offset)
 {
 	int i;
 
-	i = length_dotted(message, offset);
+	i = __length_dotted(message, offset);
 	if (i < 0)
 		return i;
 
@@ -375,11 +407,11 @@ int length_question(unsigned char *message, int offset)
 #endif
 
 #ifdef L_encodea
-int encode_answer(struct resolv_answer *a, unsigned char *dest, int maxlen)
+int __encode_answer(struct resolv_answer *a, unsigned char *dest, int maxlen)
 {
 	int i;
 
-	i = encode_dotted(a->dotted, dest, maxlen);
+	i = __encode_dotted(a->dotted, dest, maxlen);
 	if (i < 0)
 		return i;
 
@@ -406,13 +438,13 @@ int encode_answer(struct resolv_answer *a, unsigned char *dest, int maxlen)
 #endif
 
 #ifdef L_decodea
-int decode_answer(unsigned char *message, int offset,
+int __decode_answer(unsigned char *message, int offset,
 				  struct resolv_answer *a)
 {
 	char temp[256];
 	int i;
 
-	i = decode_dotted(message, offset, temp, sizeof(temp));
+	i = __decode_dotted(message, offset, temp, sizeof(temp));
 	if (i < 0)
 		return i;
 
@@ -438,7 +470,7 @@ int decode_answer(unsigned char *message, int offset,
 #endif
 
 #ifdef L_encodep
-int encode_packet(struct resolv_header *h,
+int __encode_packet(struct resolv_header *h,
 	struct resolv_question **q,
 	struct resolv_answer **an,
 	struct resolv_answer **ns,
@@ -448,7 +480,7 @@ int encode_packet(struct resolv_header *h,
 	int i, total = 0;
 	int j;
 
-	i = encode_header(h, dest, maxlen);
+	i = __encode_header(h, dest, maxlen);
 	if (i < 0)
 		return i;
 
@@ -457,7 +489,7 @@ int encode_packet(struct resolv_header *h,
 	total += i;
 
 	for (j = 0; j < h->qdcount; j++) {
-		i = encode_question(q[j], dest, maxlen);
+		i = __encode_question(q[j], dest, maxlen);
 		if (i < 0)
 			return i;
 		dest += i;
@@ -466,7 +498,7 @@ int encode_packet(struct resolv_header *h,
 	}
 
 	for (j = 0; j < h->ancount; j++) {
-		i = encode_answer(an[j], dest, maxlen);
+		i = __encode_answer(an[j], dest, maxlen);
 		if (i < 0)
 			return i;
 		dest += i;
@@ -474,7 +506,7 @@ int encode_packet(struct resolv_header *h,
 		total += i;
 	}
 	for (j = 0; j < h->nscount; j++) {
-		i = encode_answer(ns[j], dest, maxlen);
+		i = __encode_answer(ns[j], dest, maxlen);
 		if (i < 0)
 			return i;
 		dest += i;
@@ -482,7 +514,7 @@ int encode_packet(struct resolv_header *h,
 		total += i;
 	}
 	for (j = 0; j < h->arcount; j++) {
-		i = encode_answer(ar[j], dest, maxlen);
+		i = __encode_answer(ar[j], dest, maxlen);
 		if (i < 0)
 			return i;
 		dest += i;
@@ -495,14 +527,14 @@ int encode_packet(struct resolv_header *h,
 #endif
 
 #ifdef L_decodep
-int decode_packet(unsigned char *data, struct resolv_header *h)
+int __decode_packet(unsigned char *data, struct resolv_header *h)
 {
-	return decode_header(data, h);
+	return __decode_header(data, h);
 }
 #endif
 
 #ifdef L_formquery
-int form_query(int id, const char *name, int type, unsigned char *packet,
+int __form_query(int id, const char *name, int type, unsigned char *packet,
 			   int maxlen)
 {
 	struct resolv_header h;
@@ -517,11 +549,11 @@ int form_query(int id, const char *name, int type, unsigned char *packet,
 	q.qtype = type;
 	q.qclass = C_IN; /* CLASS_IN */
 
-	i = encode_header(&h, packet, maxlen);
+	i = __encode_header(&h, packet, maxlen);
 	if (i < 0)
 		return i;
 
-	j = encode_question(&q, packet + i, maxlen - i);
+	j = __encode_question(&q, packet + i, maxlen - i);
 	if (j < 0)
 		return j;
 
@@ -532,7 +564,6 @@ int form_query(int id, const char *name, int type, unsigned char *packet,
 #ifdef L_dnslookup
 
 #ifdef __UCLIBC_HAS_THREADS__
-#include <pthread.h>
 static pthread_mutex_t mylock = PTHREAD_MUTEX_INITIALIZER;
 # define LOCK	pthread_mutex_lock(&mylock)
 # define UNLOCK	pthread_mutex_unlock(&mylock);
@@ -541,30 +572,29 @@ static pthread_mutex_t mylock = PTHREAD_MUTEX_INITIALIZER;
 # define UNLOCK
 #endif
 
-/* Just for the record, having to lock dns_lookup() just for these two globals
- * is pretty lame.  Sometime I should work on making the locking a bit more
- * careful to avoid needless blocking...  */
+/* Just for the record, having to lock __dns_lookup() just for these two globals
+ * is pretty lame.  I think these two variables can probably be de-global-ized, 
+ * which should eliminate the need for doing locking here...  Needs a closer 
+ * look anyways. */
 static int ns=0, id=1;
 
-int dns_lookup(const char *name, int type, int nscount, char **nsip,
+int __dns_lookup(const char *name, int type, int nscount, char **nsip,
 			   unsigned char **outpacket, struct resolv_answer *a)
 {
-	int i, j, len, fd, pos;
-	struct sockaddr_in sa;
-#ifdef __UCLIBC_HAS_IPV6__
-	struct sockaddr_in6 sa6;
-#endif /* __UCLIBC_HAS_IPV6__ */
+	int i, j, len, fd, pos, rc;
 	struct timeval tv;
 	fd_set fds;
 	struct resolv_header h;
 	struct resolv_question q;
 	int retries = 0;
 	unsigned char * packet = malloc(PACKETSZ);
-	char * lookup = malloc(MAXDNAME);
+	char *dns, *lookup = malloc(MAXDNAME);
 	int variant = 0;
+	struct sockaddr_in sa;
 #ifdef __UCLIBC_HAS_IPV6__
 	int v6;
-#endif /* __UCLIBC_HAS_IPV6__ */
+	struct sockaddr_in6 sa6;
+#endif
 
 	fd = -1;
 
@@ -575,85 +605,88 @@ int dns_lookup(const char *name, int type, int nscount, char **nsip,
 
 	LOCK;
 	ns %= nscount;
+	UNLOCK;
 
 	while (retries++ < MAX_RETRIES) {
-#ifdef __UCLIBC_HAS_IPV6__
-		v6 = (inet_pton(AF_INET6, nsip[ns], &sa6.sin6_addr) > 0);
-#endif /* __UCLIBC_HAS_IPV6__ */
-
 		if (fd != -1)
 			close(fd);
 
-#ifndef __UCLIBC_HAS_IPV6__
-		fd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
-#else /* __UCLIBC_HAS_IPV6__ */
-		fd = socket(v6 ? AF_INET6 : AF_INET, SOCK_DGRAM, IPPROTO_UDP);
-#endif /* __UCLIBC_HAS_IPV6__ */
-
-		if (fd == -1)
-			goto fail;
-
 		memset(packet, 0, PACKETSZ);
 
 		memset(&h, 0, sizeof(h));
-		h.id = ++id;
+
+		/* Mess with globals while under lock */
+		LOCK;
+		++id;
+		id &= 0xffff;
+		h.id = id;
+		dns = nsip[ns];
+		UNLOCK;
+
 		h.qdcount = 1;
 		h.rd = 1;
 
 		DPRINTF("encoding header\n", h.rd);
 
-		i = encode_header(&h, packet, PACKETSZ);
+		i = __encode_header(&h, packet, PACKETSZ);
 		if (i < 0)
 			goto fail;
 
 		strncpy(lookup,name,MAXDNAME);
-		if (variant < searchdomains && strchr(lookup, '.') == NULL)
+		BIGLOCK;
+		if (variant < __searchdomains && strchr(lookup, '.') == NULL)
 		{
 		    strncat(lookup,".", MAXDNAME);
-		    strncat(lookup,searchdomain[variant], MAXDNAME);
+		    strncat(lookup,__searchdomain[variant], MAXDNAME);
 		}
+		BIGUNLOCK;
 		DPRINTF("lookup name: %s\n", lookup);
 		q.dotted = (char *)lookup;
 		q.qtype = type;
 		q.qclass = C_IN; /* CLASS_IN */
 
-		j = encode_question(&q, packet+i, PACKETSZ-i);
+		j = __encode_question(&q, packet+i, PACKETSZ-i);
 		if (j < 0)
 			goto fail;
 
 		len = i + j;
 
 		DPRINTF("On try %d, sending query to port %d of machine %s\n",
-				retries, NAMESERVER_PORT, nsip[ns]);
+				retries, NAMESERVER_PORT, dns);
 
-#ifndef __UCLIBC_HAS_IPV6__
-		sa.sin_family = AF_INET;
-		sa.sin_port = htons(NAMESERVER_PORT);
-		sa.sin_addr.s_addr = inet_addr(nsip[ns]);
-#else /* __UCLIBC_HAS_IPV6__ */
-		if (v6) {
-			sa6.sin6_family = AF_INET6;
-			sa6.sin6_port = htons(NAMESERVER_PORT);
-			/* sa6.sin6_addr is already here */
-		} else {
-			sa.sin_family = AF_INET;
-			sa.sin_port = htons(NAMESERVER_PORT);
-			sa.sin_addr.s_addr = inet_addr(nsip[ns]);
+#ifdef __UCLIBC_HAS_IPV6__
+		v6 = inet_pton(AF_INET6, dns, &sa6.sin6_addr) > 0;
+		fd = socket(v6 ? AF_INET6 : AF_INET, SOCK_DGRAM, IPPROTO_UDP);
+#else
+		fd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
+#endif
+		if (fd < 0) {
+		    continue;
 		}
-#endif /* __UCLIBC_HAS_IPV6__ */
 
-#ifndef __UCLIBC_HAS_IPV6__
-		if (connect(fd, (struct sockaddr *) &sa, sizeof(sa)) == -1) {
-#else /* __UCLIBC_HAS_IPV6__ */
-		if (connect(fd, (struct sockaddr *) (v6 ? &sa6 : &sa), 
-			    v6 ? sizeof(sa6) : sizeof(sa)) == -1) {
-#endif /* __UCLIBC_HAS_IPV6__ */
-			if (errno == ENETUNREACH) {
-				/* routing error, presume not transient */
-				goto tryall;
-			} else
-				/* retry */
-				continue;
+		/* Connect to the UDP socket so that asyncronous errors are returned */		 
+#ifdef __UCLIBC_HAS_IPV6__
+		if (v6) {
+		    sa6.sin6_family = AF_INET6;
+		    sa6.sin6_port = htons(NAMESERVER_PORT);
+		    /* sa6.sin6_addr is already here */
+		    rc = connect(fd, (struct sockaddr *) &sa6, sizeof(sa6));
+		} else {
+#endif
+		    sa.sin_family = AF_INET;
+		    sa.sin_port = htons(NAMESERVER_PORT);
+		    sa.sin_addr.s_addr = inet_addr(dns);
+		    rc = connect(fd, (struct sockaddr *) &sa, sizeof(sa));
+#ifdef __UCLIBC_HAS_IPV6__
+		}
+#endif
+		if (rc < 0) {
+		    if (errno == ENETUNREACH) {
+			/* routing error, presume not transient */
+			goto tryall;
+		    } else
+			/* retry */
+			continue;
 		}
 
 		DPRINTF("Transmitting packet of length %d, id=%d, qr=%d\n",
@@ -674,17 +707,23 @@ int dns_lookup(const char *name, int type, int nscount, char **nsip,
 		}
 
 		i = recv(fd, packet, 512, 0);
-		if (i < HFIXEDSZ)
+		if (i < HFIXEDSZ) {
 			/* too short ! */
 			goto again;
+		}
 
-		decode_header(packet, &h);
+		__decode_header(packet, &h);
 
 		DPRINTF("id = %d, qr = %d\n", h.id, h.qr);
 
-		if ((h.id != id) || (!h.qr))
+		LOCK;
+		if ((h.id != id) || (!h.qr)) {
+			UNLOCK;
 			/* unsolicited */
 			goto again;
+		}
+		UNLOCK;
+
 
 		DPRINTF("Got response %s\n", "(i think)!");
 		DPRINTF("qrcount=%d,ancount=%d,nscount=%d,arcount=%d\n",
@@ -701,7 +740,7 @@ int dns_lookup(const char *name, int type, int nscount, char **nsip,
 
 		for (j = 0; j < h.qdcount; j++) {
 			DPRINTF("Skipping question %d at %d\n", j, pos);
-			i = length_question(packet, pos);
+			i = __length_question(packet, pos);
 			DPRINTF("Length of question %d is %d\n", j, i);
 			if (i < 0)
 				goto again;
@@ -711,7 +750,7 @@ int dns_lookup(const char *name, int type, int nscount, char **nsip,
 
 		for (j=0;j<h.ancount;j++)
 		{
-		    i = decode_answer(packet, pos, a);
+		    i = __decode_answer(packet, pos, a);
 
 		    if (i<0) {
 			DPRINTF("failed decode %d\n", i);
@@ -736,30 +775,44 @@ int dns_lookup(const char *name, int type, int nscount, char **nsip,
 		else
 			free(packet);
 		free(lookup);
-		UNLOCK;
 		return (0);				/* success! */
 
 	  tryall:
 		/* if there are other nameservers, give them a go,
 		   otherwise return with error */
-		variant = 0;
-		if (retries >= nscount*(searchdomains+1))
-		    goto fail;
+		{
+		    int sdomains;
+
+		    BIGLOCK;
+		    sdomains=__searchdomains;
+		    BIGUNLOCK;
+		    variant = 0;
+		    if (retries >= nscount*(sdomains+1))
+			goto fail;
+		}
 
 	  again:
 		/* if there are searchdomains, try them or fallback as passed */
-		if (variant < searchdomains) {
-		    /* next search */
-		    variant++;
-		} else {
-		    /* next server, first search */
-		    ns = (ns + 1) % nscount;
-		    variant = 0;
+		{
+		    int sdomains;
+		    BIGLOCK;
+		    sdomains=__searchdomains;
+		    BIGUNLOCK;
+
+		    if (variant < sdomains) {
+			/* next search */
+			variant++;
+		    } else {
+			/* next server, first search */
+			LOCK;
+			ns = (ns + 1) % nscount;
+			UNLOCK;
+			variant = 0;
+		    }
 		}
 	}
 
 fail:
-	UNLOCK;
 	if (fd != -1)
 	    close(fd);
 	if (lookup)
@@ -772,17 +825,20 @@ fail:
 
 #ifdef L_opennameservers
 
-int nameservers;
-char * nameserver[MAX_SERVERS];
-int searchdomains;
-char * searchdomain[MAX_SEARCH];
+int __nameservers;
+char * __nameserver[MAX_SERVERS];
+int __searchdomains;
+char * __searchdomain[MAX_SEARCH];
+#ifdef __UCLIBC_HAS_THREADS__
+pthread_mutex_t __resolv_lock = PTHREAD_MUTEX_INITIALIZER;
+#endif
 
 /*
  *	we currently read formats not quite the same as that on normal
  *	unix systems, we can have a list of nameservers after the keyword.
  */
 
-int open_nameservers()
+int __open_nameservers()
 {
 	FILE *fp;
 	int i;
@@ -790,8 +846,11 @@ int open_nameservers()
 	char szBuffer[128], *p, *argv[RESOLV_ARGS];
 	int argc;
 
-	if (nameservers > 0) 
+	BIGLOCK;
+	if (__nameservers > 0) { 
+	    BIGUNLOCK;
 	    return 0;
+	}
 
 	if ((fp = fopen("/etc/resolv.conf", "r")) ||
 			(fp = fopen("/etc/config/resolv.conf", "r"))) {
@@ -812,20 +871,20 @@ int open_nameservers()
 			}
 
 			if (strcmp(argv[0], "nameserver") == 0) {
-				for (i = 1; i < argc && nameservers < MAX_SERVERS; i++) {
-					nameserver[nameservers++] = strdup(argv[i]);
+				for (i = 1; i < argc && __nameservers < MAX_SERVERS; i++) {
+					__nameserver[__nameservers++] = strdup(argv[i]);
 					DPRINTF("adding nameserver %s\n", argv[i]);
 				}
 			}
 
 			/* domain and search are mutually exclusive, the last one wins */
 			if (strcmp(argv[0],"domain")==0 || strcmp(argv[0],"search")==0) {
-				while (searchdomains > 0) {
-					free(searchdomain[--searchdomains]);
-					searchdomain[searchdomains] = NULL;
+				while (__searchdomains > 0) {
+					free(__searchdomain[--__searchdomains]);
+					__searchdomain[__searchdomains] = NULL;
 				}
-				for (i=1; i < argc && searchdomains < MAX_SEARCH; i++) {
-					searchdomain[searchdomains++] = strdup(argv[i]);
+				for (i=1; i < argc && __searchdomains < MAX_SEARCH; i++) {
+					__searchdomain[__searchdomains++] = strdup(argv[i]);
 					DPRINTF("adding search %s\n", argv[i]);
 				}
 			}
@@ -834,7 +893,8 @@ int open_nameservers()
 	} else {
 	    DPRINTF("failed to open %s\n", "resolv.conf");
 	}
-	DPRINTF("nameservers = %d\n", nameservers);
+	DPRINTF("nameservers = %d\n", __nameservers);
+	BIGUNLOCK;
 	return 0;
 }
 #endif
@@ -842,16 +902,18 @@ int open_nameservers()
 
 #ifdef L_closenameservers
 
-void close_nameservers(void)
+void __close_nameservers(void)
 {
-	while (nameservers > 0) {
-		free(nameserver[--nameservers]);
-		nameserver[nameservers] = NULL;
+	BIGLOCK;
+	while (__nameservers > 0) {
+		free(__nameserver[--__nameservers]);
+		__nameserver[__nameservers] = NULL;
 	}
-	while (searchdomains > 0) {
-		free(searchdomain[--searchdomains]);
-		searchdomain[searchdomains] = NULL;
+	while (__searchdomains > 0) {
+		free(__searchdomain[--__searchdomains]);
+		__searchdomain[__searchdomains] = NULL;
 	}
+	BIGUNLOCK;
 }
 #endif
 
@@ -862,7 +924,7 @@ struct hostent *gethostbyname(const char *name)
 	static struct hostent h;
 	static char buf[sizeof(struct in_addr) +
 			sizeof(struct in_addr *)*2 +
-			256/*namebuffer*/ + 32/* margin */];
+			sizeof(char *)*(ALIAS_DIM) + 256/*namebuffer*/ + 32/* margin */];
 	struct hostent *hp;
 
 	gethostbyname_r(name, &h, buf, sizeof(buf), &hp, &h_errno);
@@ -873,14 +935,6 @@ struct hostent *gethostbyname(const char *name)
 
 #ifdef L_gethostbyname2
 
-#ifdef __UCLIBC_HAS_IPV6__
-/* TBD: Not the right place for defining these, I guess */
-const struct in6_addr in6addr_any =
-	{ { { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 } } };
-const struct in6_addr in6addr_loopback =
-	{ { { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1 } } };                                    
-#endif /* __UCLIBC_HAS_IPV6__ */
-
 struct hostent *gethostbyname2(const char *name, int family)
 {
 #ifndef __UCLIBC_HAS_IPV6__
@@ -889,7 +943,7 @@ struct hostent *gethostbyname2(const char *name, int family)
 	static struct hostent h;
 	static char buf[sizeof(struct in6_addr) +
 			sizeof(struct in6_addr *)*2 +
-			256/*namebuffer*/ + 32/* margin */];
+			sizeof(char *)*(ALIAS_DIM) + 256/*namebuffer*/ + 32/* margin */];
 	struct hostent *hp;
 
 	gethostbyname2_r(name, family, &h, buf, sizeof(buf), &hp, &h_errno);
@@ -909,22 +963,14 @@ struct netent * getnetbyname(const char * name)
 
 
 #ifdef L_res_init
-struct __res_state * __res;
-
-#ifndef _res
-#define _res (*__res_state())
-#endif
+struct __res_state _res;
 
 int res_init(void)
 {
-	struct __res_state *rp = __res;
-	if(!__res) {
-		rp = (struct __res_state *) malloc(sizeof(struct __res_state));
-		memset(rp, 0, sizeof(struct __res_state));
-		__res = rp;
-	}	
+	struct __res_state *rp = &(_res);
 
-	(void) open_nameservers();
+	__close_nameservers();
+	__open_nameservers();
 	rp->retrans = RES_TIMEOUT;
 	rp->retry = 4;
 	rp->options = RES_INIT;
@@ -940,43 +986,33 @@ int res_init(void)
 	/** rp->rhook = NULL; **/
 	/** rp->_u._ext.nsinit = 0; **/
 
-	if(searchdomains) {
+	BIGLOCK;
+	if(__searchdomains) {
 		int i;
-		for(i=0; i<searchdomains; i++) {
-			rp->dnsrch[i] = searchdomain[i];
+		for(i=0; i<__searchdomains; i++) {
+			rp->dnsrch[i] = __searchdomain[i];
 		}
 	}
 
-	if(nameservers) {
+	if(__nameservers) {
 		int i;
 		struct in_addr a;
-		for(i=0; i<nameservers; i++) {
-			if (inet_aton(nameserver[i], &a)) {
+		for(i=0; i<__nameservers; i++) {
+			if (inet_aton(__nameserver[i], &a)) {
 				rp->nsaddr_list[i].sin_addr = a;
 				rp->nsaddr_list[i].sin_family = AF_INET;
 				rp->nsaddr_list[i].sin_port = htons(NAMESERVER_PORT);
 			}
 		}
 	}
-	rp->nscount = nameservers;
+	rp->nscount = __nameservers;
+	BIGUNLOCK;
 
 	return(0);
 }
 
-struct __res_state * __res_state (void)
-{
-	if(!__res) {
-		res_init();
-	}
-	return __res;
-}
-
 void res_close( void )
 {
-	if(__res) {
-		free(__res);
-		__res = NULL;
-	}
 	return;
 }
 
@@ -992,18 +1028,24 @@ void res_close( void )
 int res_query(const char *dname, int class, int type,
               unsigned char *answer, int anslen)
 {
+	int i;
 	unsigned char * packet = 0;
 	struct resolv_answer a;
-	int i;
+	int __nameserversXX;
+	char ** __nameserverXX;
 
-	open_nameservers();
+	__open_nameservers();
 	
 	if (!dname || class != 1 /* CLASS_IN */)
 		return(-1);
 		
 	memset((char *) &a, '\0', sizeof(a));
 
-	i = dns_lookup(dname, type, nameservers, nameserver, &packet, &a);
+	BIGLOCK;
+	__nameserversXX=__nameservers;
+	__nameserverXX=__nameserver;
+	BIGUNLOCK;
+	i = __dns_lookup(dname, type, __nameserversXX, __nameserverXX, &packet, &a);
 	
 	if (i < 0)
 		return(-1);
@@ -1023,7 +1065,6 @@ int res_query(const char *dname, int class, int type,
 }
 #endif
 
-
 #ifdef L_gethostbyaddr
 struct hostent *gethostbyaddr (const void *addr, socklen_t len, int type)
 {
@@ -1034,7 +1075,7 @@ struct hostent *gethostbyaddr (const void *addr, socklen_t len, int type)
 #else
 		sizeof(struct in6_addr) + sizeof(struct in6_addr *)*2 +
 #endif /* __UCLIBC_HAS_IPV6__ */
-		256/*namebuffer*/ + 32/* margin */];
+		sizeof(char *)*(ALIAS_DIM) + 256/*namebuffer*/ + 32/* margin */];
 	struct hostent *hp;
 
 	gethostbyaddr_r(addr, len, type, &h, buf, sizeof(buf), &hp, &h_errno);
@@ -1054,7 +1095,7 @@ void __open_etc_hosts(FILE **fp)
 	return;
 }
 
-int read_etc_hosts_r(FILE * fp, const char * name, int type,
+int __read_etc_hosts_r(FILE * fp, const char * name, int type,
 		     enum etc_hosts_action action,
 		     struct hostent * result_buf,
 		     char * buf, size_t buflen,
@@ -1068,11 +1109,16 @@ int read_etc_hosts_r(FILE * fp, const char * name, int type,
 	struct in6_addr	**addr_list6=NULL;
 #endif /* __UCLIBC_HAS_IPV6__ */
 	char					*cp;
-#define		 MAX_ALIAS		5
-	char					*alias[MAX_ALIAS];
+	char					**alias;
 	int						aliases, i;
 	int		ret=HOST_NOT_FOUND;
 
+	if (buflen < sizeof(char *)*(ALIAS_DIM))
+		return ERANGE;
+	alias=(char **)buf;
+	buf+=sizeof(char **)*(ALIAS_DIM);
+	buflen-=sizeof(char **)*(ALIAS_DIM);
+
 	if (action!=GETHOSTENT) {
 #ifdef __UCLIBC_HAS_IPV6__
 		char *p=buf;
@@ -1109,6 +1155,7 @@ int read_etc_hosts_r(FILE * fp, const char * name, int type,
 			buf=p;
 		}
 #endif /* __UCLIBC_HAS_IPV6__ */
+
 		if (buflen < 80)
 			return ERANGE;
 
@@ -1132,11 +1179,12 @@ int read_etc_hosts_r(FILE * fp, const char * name, int type,
 				*cp++ = '\0';
 			if (!*cp)
 				continue;
-			if (aliases < MAX_ALIAS)
+			if (aliases < (2+MAX_ALIASES))
 				alias[aliases++] = cp;
 			while (*cp && !isspace(*cp))
 				cp++;
 		}
+		alias[aliases] = 0;
 
 		if (aliases < 2)
 			continue; /* syntax error really */
@@ -1164,6 +1212,7 @@ int read_etc_hosts_r(FILE * fp, const char * name, int type,
 			result_buf->h_addrtype = AF_INET;
 			result_buf->h_length = sizeof(*in);
 			result_buf->h_addr_list = (char**) addr_list;
+			result_buf->h_aliases = alias + 2;
 			*result=result_buf;
 			ret=NETDB_SUCCESS;
 #ifdef __UCLIBC_HAS_IPV6__
@@ -1175,6 +1224,7 @@ int read_etc_hosts_r(FILE * fp, const char * name, int type,
 			result_buf->h_addrtype = AF_INET6;
 			result_buf->h_length = sizeof(*in6);
 			result_buf->h_addr_list = (char**) addr_list6;
+			result_buf->h_aliases = alias + 2;
 			*result=result_buf;
 			ret=NETDB_SUCCESS;
 #endif /* __UCLIBC_HAS_IPV6__ */
@@ -1200,7 +1250,6 @@ int read_etc_hosts_r(FILE * fp, const char * name, int type,
 #ifdef L_gethostent
 
 #ifdef __UCLIBC_HAS_THREADS__
-#include <pthread.h>
 static pthread_mutex_t mylock = PTHREAD_MUTEX_INITIALIZER;
 # define LOCK	pthread_mutex_lock(&mylock)
 # define UNLOCK	pthread_mutex_unlock(&mylock);
@@ -1239,6 +1288,7 @@ struct hostent *gethostent (void)
 #else
 	    sizeof(struct in6_addr) + sizeof(struct in6_addr *)*2 +
 #endif /* __UCLIBC_HAS_IPV6__ */
+		sizeof(char *)*(ALIAS_DIM) +
 	    80/*namebuffer*/ + 2/* margin */];
     struct hostent *host;
 
@@ -1251,7 +1301,7 @@ struct hostent *gethostent (void)
 	}
     }
 
-    read_etc_hosts_r(__gethostent_fp, NULL, AF_INET, GETHOSTENT, 
+    __read_etc_hosts_r(__gethostent_fp, NULL, AF_INET, GETHOSTENT, 
 		   &h, buf, sizeof(buf), &host, &h_errno);
     if (__stay_open==0) {
 	fclose(__gethostent_fp);
@@ -1263,19 +1313,19 @@ struct hostent *gethostent (void)
 
 #ifdef L_get_hosts_byname_r
 
-int get_hosts_byname_r(const char * name, int type,
+int __get_hosts_byname_r(const char * name, int type,
 			    struct hostent * result_buf,
 			    char * buf, size_t buflen,
 			    struct hostent ** result,
 			    int * h_errnop)
 {
-	return(read_etc_hosts_r(NULL, name, type, GET_HOSTS_BYNAME, result_buf, buf, buflen, result, h_errnop));
+	return(__read_etc_hosts_r(NULL, name, type, GET_HOSTS_BYNAME, result_buf, buf, buflen, result, h_errnop));
 }
 #endif
 
 #ifdef L_get_hosts_byaddr_r
 
-int get_hosts_byaddr_r(const char * addr, int len, int type,
+int __get_hosts_byaddr_r(const char * addr, int len, int type,
 			    struct hostent * result_buf,
 			    char * buf, size_t buflen,
 			    struct hostent ** result,
@@ -1304,7 +1354,8 @@ int get_hosts_byaddr_r(const char * addr, int len, int type,
 
 	inet_ntop(type, addr, ipaddr, sizeof(ipaddr));
 
-	return(read_etc_hosts_r(NULL, ipaddr, type, GET_HOSTS_BYADDR, result_buf, buf, buflen, result, h_errnop));
+	return(__read_etc_hosts_r(NULL, ipaddr, type, GET_HOSTS_BYADDR, 
+		    result_buf, buf, buflen, result, h_errnop));
 }
 #endif
 
@@ -1516,15 +1567,17 @@ int gethostbyname_r(const char * name,
 	struct resolv_answer a;
 	int i;
 	int nest = 0;
+	int __nameserversXX;
+	char ** __nameserverXX;
 
-	open_nameservers();
+	__open_nameservers();
 
 	*result=NULL;
 	if (!name)
 		return EINVAL;
 
 	/* do /etc/hosts first */
-	if ((i=get_hosts_byname_r(name, AF_INET, result_buf,
+	if ((i=__get_hosts_byname_r(name, AF_INET, result_buf,
 				  buf, buflen, result, h_errnop))==0)
 		return i;
 	switch (*h_errnop) {
@@ -1575,11 +1628,15 @@ int gethostbyname_r(const char * name,
 
 	for (;;) {
 
-		i = dns_lookup(buf, T_A, nameservers, nameserver, &packet, &a);
+	BIGLOCK;
+	__nameserversXX=__nameservers;
+	__nameserverXX=__nameserver;
+	BIGUNLOCK;
+		i = __dns_lookup(buf, T_A, __nameserversXX, __nameserverXX, &packet, &a);
 
 		if (i < 0) {
 			*h_errnop = HOST_NOT_FOUND;
-			DPRINTF("dns_lookup\n");
+			DPRINTF("__dns_lookup\n");
 			return TRY_AGAIN;
 		}
 
@@ -1588,12 +1645,12 @@ int gethostbyname_r(const char * name,
 
 		if (a.atype == T_CNAME) {		/* CNAME */
 			DPRINTF("Got a CNAME in gethostbyname()\n");
-			i = decode_dotted(packet, a.rdoffset, buf, buflen);
+			i = __decode_dotted(packet, a.rdoffset, buf, buflen);
 			free(packet);
 
 			if (i < 0) {
 				*h_errnop = NO_RECOVERY;
-				DPRINTF("decode_dotted\n");
+				DPRINTF("__decode_dotted\n");
 				return -1;
 			}
 			if (++nest > MAX_RECURSE) {
@@ -1624,16 +1681,6 @@ int gethostbyname_r(const char * name,
 
 #ifdef L_gethostbyname2_r
 
-#ifdef __UCLIBC_HAS_IPV6__
-/* TBD: Not the right place for defining these, I guess */
-/*
-const struct in6_addr in6addr_any =
-	{ { { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 } } };
-const struct in6_addr in6addr_loopback =
-	{ { { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1 } } };
-*/
-#endif /* __UCLIBC_HAS_IPV6__ */
-
 int gethostbyname2_r(const char *name, int family,
 			    struct hostent * result_buf,
 			    char * buf, size_t buflen,
@@ -1649,6 +1696,8 @@ int gethostbyname2_r(const char *name, int family,
 	struct resolv_answer a;
 	int i;
 	int nest = 0;
+	int __nameserversXX;
+	char ** __nameserverXX;
 
 	if (family == AF_INET)
 		return gethostbyname_r(name, result_buf, buf, buflen, result, h_errnop);
@@ -1656,14 +1705,14 @@ int gethostbyname2_r(const char *name, int family,
 	if (family != AF_INET6)
 		return EINVAL;
 		
-	open_nameservers();
+	__open_nameservers();
 
 	*result=NULL;
 	if (!name)
 		return EINVAL;
 
 	/* do /etc/hosts first */
-	if ((i=get_hosts_byname_r(name, family, result_buf,
+	if ((i=__get_hosts_byname_r(name, family, result_buf,
 				  buf, buflen, result, h_errnop))==0)
 		return i;
 	switch (*h_errnop) {
@@ -1708,8 +1757,12 @@ int gethostbyname2_r(const char *name, int family,
 	}
 
 	for (;;) {
+	BIGLOCK;
+	__nameserversXX=__nameservers;
+	__nameserverXX=__nameserver;
+	BIGUNLOCK;
 
-		i = dns_lookup(buf, T_AAAA, nameservers, nameserver, &packet, &a);
+		i = __dns_lookup(buf, T_AAAA, __nameserversXX, __nameserverXX, &packet, &a);
 
 		if (i < 0) {
 			*h_errnop = HOST_NOT_FOUND;
@@ -1721,7 +1774,7 @@ int gethostbyname2_r(const char *name, int family,
 
 		if (a.atype == T_CNAME) {		/* CNAME */
 			DPRINTF("Got a CNAME in gethostbyname()\n");
-			i = decode_dotted(packet, a.rdoffset, buf, buflen);
+			i = __decode_dotted(packet, a.rdoffset, buf, buflen);
 			free(packet);
 
 			if (i < 0) {
@@ -1774,6 +1827,8 @@ int gethostbyaddr_r (const void *addr, socklen_t len, int type,
 	struct resolv_answer a;
 	int i;
 	int nest = 0;
+	int __nameserversXX;
+	char ** __nameserverXX;
 
 	*result=NULL;
 	if (!addr)
@@ -1795,7 +1850,7 @@ int gethostbyaddr_r (const void *addr, socklen_t len, int type,
 	}
 
 	/* do /etc/hosts first */
-	if ((i=get_hosts_byaddr_r(addr, len, type, result_buf,
+	if ((i=__get_hosts_byaddr_r(addr, len, type, result_buf,
 				  buf, buflen, result, h_errnop))==0)
 		return i;
 	switch (*h_errnop) {
@@ -1806,7 +1861,7 @@ int gethostbyaddr_r (const void *addr, socklen_t len, int type,
 			return i;
 	}
 
-	open_nameservers();
+	__open_nameservers();
 
 #ifdef __UCLIBC_HAS_IPV6__
 	qp=buf;
@@ -1876,7 +1931,11 @@ int gethostbyaddr_r (const void *addr, socklen_t len, int type,
 
 	for (;;) {
 
-		i = dns_lookup(buf, T_PTR, nameservers, nameserver, &packet, &a);
+	BIGLOCK;
+	__nameserversXX=__nameservers;
+	__nameserverXX=__nameserver;
+	BIGUNLOCK;
+		i = __dns_lookup(buf, T_PTR, __nameserversXX, __nameserverXX, &packet, &a);
 
 		if (i < 0) {
 			*h_errnop = HOST_NOT_FOUND;
@@ -1888,7 +1947,7 @@ int gethostbyaddr_r (const void *addr, socklen_t len, int type,
 
 		if (a.atype == T_CNAME) {		/* CNAME */
 			DPRINTF("Got a CNAME in gethostbyaddr()\n");
-			i = decode_dotted(packet, a.rdoffset, buf, buflen);
+			i = __decode_dotted(packet, a.rdoffset, buf, buflen);
 			free(packet);
 
 			if (i < 0) {
@@ -1901,7 +1960,7 @@ int gethostbyaddr_r (const void *addr, socklen_t len, int type,
 			}
 			continue;
 		} else if (a.atype == T_PTR) {	/* ADDRESS */
-			i = decode_dotted(packet, a.rdoffset, buf, buflen);
+			i = __decode_dotted(packet, a.rdoffset, buf, buflen);
 			free(packet);
 
 			result_buf->h_name = buf;
