@@ -17,34 +17,45 @@
  *
  */
 
+#include <features.h>
 #include <stdlib.h>
 #include <unistd.h>
 #include <fcntl.h>
-#include <pwd.h>
-#include <shadow.h>
+#include "config.h"
 
-#define PWD_BUFFER_SIZE 256
+#ifdef __UCLIBC_HAS_THREADS__
+#include <pthread.h>
+static pthread_mutex_t mylock = PTHREAD_MUTEX_INITIALIZER;
+# define LOCK   pthread_mutex_lock(&mylock)
+# define UNLOCK pthread_mutex_unlock(&mylock);
+#else       
+# define LOCK
+# define UNLOCK
+#endif      
 
 int getspuid_r (uid_t uid, struct spwd *spwd,
 	char *buff, size_t buflen, struct spwd **crap)
 {
-	char pwd_buff[PWD_BUFFER_SIZE];
-	struct passwd password;
+    char pwd_buff[PWD_BUFFER_SIZE];
+    struct passwd password;
 
-	if (getpwuid_r(uid, &password, pwd_buff, PWD_BUFFER_SIZE, NULL) < 0)
-		return -1;
+    if (getpwuid_r(uid, &password, pwd_buff,  sizeof(pwd_buff), NULL) < 0)
+	return -1;
 
-	return getspnam_r(password.pw_name, spwd, buff, buflen, crap);
+    return getspnam_r(password.pw_name, spwd, buff, buflen, crap);
 }
 
 struct spwd *getspuid(uid_t uid)
 {
-	static char line_buff[PWD_BUFFER_SIZE];
-	static struct spwd spwd;
+    static char line_buff[PWD_BUFFER_SIZE];
+    static struct spwd spwd;
 
-	if (getspuid_r(uid, &spwd, line_buff, PWD_BUFFER_SIZE, NULL) != -1) {
-		return &spwd;
-	}
-	return NULL;
+    LOCK;
+    if (getspuid_r(uid, &spwd, line_buff, sizeof(line_buff), NULL) != -1) {
+	UNLOCK;
+	return &spwd;
+    }
+    UNLOCK;
+    return NULL;
 }
 
